from math import radians
from typing import Dict
from casadi import SX, nlpsol, vertcat
import numpy as np


from trip_kinematics.KinematicGroup import KinematicGroup
from trip_kinematics.Transformation import Transformation
from trip_kinematics.Robot import Robot
from trip_kinematics.Utility import hom_translation_matrix, x_axis_rotation_matrix
from trip_kinematics.Utility import y_axis_rotation_matrix, z_axis_rotation_matrix
from trip_kinematics.Utility import hom_rotation, get_translation


def sphere_centers(r_x, r_y, r_z):
    """Generates the centers of the spheres used for the geometric closure equation

    Args:
        r_x (float): gimbal joint state rx
        r_y (float): gimbal joint state ry
        r_z (float): gimbal joint state rz

    Returns:
        numpy array, numpy array: Two 3D arrays containing the sphere centers
    """
    a_ccs_p_trans_m = hom_translation_matrix(
        t_x=0.265, t_y=0, t_z=0.014)
    a_ccs_p_rot_m = hom_rotation(x_axis_rotation_matrix(r_x) @
                                 y_axis_rotation_matrix(r_y) @
                                 z_axis_rotation_matrix(r_z))
    a_p_sph_1_2 = hom_translation_matrix(
        t_x=0.015, t_y=0.029, t_z=-0.0965)
    a_p_sph_2_2 = hom_translation_matrix(
        t_x=0.015, t_y=-0.029, t_z=-0.0965)

    a_ccs_ = a_ccs_p_trans_m @ a_ccs_p_rot_m
    a_c1 = a_ccs_ @ a_p_sph_1_2
    a_c2 = a_ccs_ @ a_p_sph_2_2

    return get_translation(a_c1), get_translation(a_c2)


def intersection_left(theta):
    """calculates the desired sphere intersection point based on the left swing joint

    Args:
        theta (float): angle of the left swing joint

    Returns:
        numpy array: the sphere intersection
    """
    a_ccs_lsm_trans = hom_translation_matrix(
        t_x=0.139807669447128, t_y=0.0549998406976098, t_z=-0.051)
    a_ccs_lsm_rot = hom_rotation(z_axis_rotation_matrix(radians(-345.0)))
    a_mcs_1_joint = hom_rotation(z_axis_rotation_matrix(theta))
    a_mcs_1_sp_1_1 = hom_translation_matrix(
        t_x=0.085, t_y=0, t_z=-0.0245)

    a_ccs_sp_1_1 = a_ccs_lsm_trans @ a_ccs_lsm_rot @ a_mcs_1_joint @ a_mcs_1_sp_1_1
    return get_translation(a_ccs_sp_1_1)


def intersection_right(theta):
    """calculates the desired sphere intersection point based on the right swing joint

    Args:
        theta (float): angle of the right swing joint

    Returns:
        numpy array: the sphere intersection
    """
    a_ccs_rsm_tran = hom_translation_matrix(
        t_x=0.139807669447128, t_y=-0.0549998406976098, t_z=-0.051)
    a_ccs_rsm_rot = hom_rotation(z_axis_rotation_matrix(radians(-15.0)))
    a_mcs_2_joint = hom_rotation(z_axis_rotation_matrix(theta))
    a_mcs_2_sp_2_1 = hom_translation_matrix(
        t_x=0.085, t_y=0, t_z=-0.0245)

    a_ccs_sp_2_1 = a_ccs_rsm_tran @ a_ccs_rsm_rot @ a_mcs_2_joint @ a_mcs_2_sp_2_1
    return get_translation(a_ccs_sp_2_1)


def swing_to_gimbal(state: Dict[str, float], tips: Dict[str, float] = None):
    """Actuated to virtual state mapping for the TriPed legss closed subchain

    Args:
        state (Dict[str, float]): actuated state of the TriPed leg closed subchain
        tips (Dict[str, float], optional): Initial state for the closure equation solver.
                                           Defaults to None in which case [0,0,0] is used.

    Returns:
        Dict[str, Dict[str, float]]: the correspdonding state of the virtual chain
    """
    x_0 = [0, 0, 0]
    if tips:
        x_0[2] = tips['rx']
        x_0[3] = tips['ry']
        x_0[4] = tips['rz']

    nlp = {'x': virtual_state, 'f': closing_equation, 'p': actuated_state}
    nlp_solver = nlpsol('swing_to_gimbal', 'ipopt', nlp, opts)
    solution = nlp_solver(x0=x_0,
                          p=[state['swing_left'], state['swing_right']])
    sol_vector = np.array(solution['x'])
    return {'gimbal_joint': {'rx': sol_vector[0][0],
                             'ry': sol_vector[1][0],
                             'rz': sol_vector[2][0]}}


def gimbal_to_swing(state: Dict[str, Dict[str, float]], tips: Dict[str, float] = None):
    """Virtual to actuated state mapping for the TriPed legss closed subchain

    Args:
        state (Dict[str, Dict[str, float]]): virtual state of the TriPed leg closed subchain
        tips (Dict[str, float], optional): Initial state for the closure equation solver.
                                           Defaults to None in which case [0,0] is used.

    Returns:
        Dict[str, float]: the correspdonding actuated state
    """
    x_0 = [0, 0]
    continuity = 0
    if tips:
        x_0[0] = tips['swing_left']
        x_0[1] = tips['swing_right']
        continuity = (x_0-actuated_state).T @ (x_0-actuated_state)

    nlp = {'x': actuated_state, 'f': closing_equation+continuity, 'p': virtual_state,
           'g': actuated_state[0]*actuated_state[1]}
    nlp_solver = nlpsol('gimbal_to_swing', 'ipopt', nlp, opts)
    solution = nlp_solver(x0=x_0,
                          p=[state['gimbal_joint']['rx'],
                             state['gimbal_joint']['ry'],
                             state['gimbal_joint']['rz']],
                          ubg=0)
    sol_vector = np.array(solution['x'])
    return {'swing_left': sol_vector[0][0], 'swing_right': sol_vector[1][0]}


theta_left = SX.sym('theta_left')
theta_right = SX.sym('theta_right')
gimbal_x = SX.sym('gimbal_x')
gimbal_y = SX.sym('gimbal_y')
gimbal_z = SX.sym('gimbal_z')

virtual_state = vertcat(gimbal_x, gimbal_y, gimbal_z)
actuated_state = vertcat(theta_left, theta_right)

opts = {'ipopt.print_level': 0, 'print_time': 0}
RADIUS = 0.11
c1, c2 = sphere_centers(r_x=gimbal_x, r_y=gimbal_y, r_z=gimbal_z)
closing_equation = (((c1-intersection_left(theta_right)).T @ (c1-intersection_left(theta_right)) -
                    RADIUS**2)**2 +
                    ((c2-intersection_right(theta_left)).T @ (c2-intersection_right(theta_left)) -
                    RADIUS**2)**2)

a_ccs_p_trans = Transformation(name='A_ccs_P_trans',
                               values={'tx': 0.265, 'tz': 0.014})
a_ccs_p_rot = Transformation(name='gimbal_joint',
                             values={'rx': 0, 'ry': 0, 'rz': 0},
                             state_variables=['rx', 'ry', 'rz'],
                             parent=a_ccs_p_trans)

closed_chain = KinematicGroup(name='closed_chain',
                              virtual_chain=[a_ccs_p_trans, a_ccs_p_rot],
                              actuated_state={
                                  'swing_left': 0, 'swing_right': 0},
                              actuated_to_virtual=swing_to_gimbal,
                              virtual_to_actuated=gimbal_to_swing)

a_p_ll = Transformation(name='A_P_LL',
                        values={'tx': 1.640, 'tz': -0.037, },
                        parent=closed_chain)
a_ll_zero = Transformation(name='zero_angle_convention',
                           values={'ry': radians(-3)},
                           parent=a_p_ll)
a_ll_zero_ll_joint = Transformation(name='extend_joint',
                                    values={'ry': 0},
                                    state_variables=['ry'],
                                    parent=a_ll_zero)
a_ll_joint_fcs = Transformation(name='A_LL_Joint_FCS',
                                values={'tx': -1.5},
                                parent=a_ll_zero_ll_joint)


triped_leg = Robot([closed_chain, a_p_ll, a_ll_zero,
                   a_ll_zero_ll_joint, a_ll_joint_fcs])
triped_leg.set_actuated_state({'swing_left': 0, 'swing_right': 0})



from trip_kinematics.KinematicGroup import *
from trip_kinematics.Transformation import *
from trip_kinematics.Robot import *
from trip_kinematics.Utility import *
from trip_kinematics.Solver import *



from typing import Dict, List, Callable
from copy import deepcopy

from trip_kinematics.Utility import identity_transformation
from trip_kinematics.Transformation import Transformation


class KinematicGroup():
    """Initializes a :py:class:`KinematicGroup` object.

    Args:
        name (str): The unique name identifying the group.
                    No two :py:class:`KinematicGroup` objects of a :py:class`Robot` should have the
                    same name
        virtual_chain (List[Transformation]): A list of :py:class:`Transformation`
                                              objects forming a serial Kinematic chain.
        actuated_state (List[Dict[str, float]]): The State of the Groups actuated joints.
        actuated_to_virtual (Callable): Maps the :py:attr:`actuated_state` to the
                                        :py:attr:`virtual_state` of the :py:attr:`virtual_chain`.
        virtual_to_actuated (Callable): Maps the :py:attr:`virtual_state` of the
                                        :py:attr:`virtual_chain` to the
                                        :py:attr:`actuated_state`.
        act_to_virt_args ([type], optional): Arguments that can be passed to
                                             :py:attr:`actuated_to_virtual` during the initial
                                             testing of the function. Defaults to None.
        virt_to_act_args ([type], optional): Arguments that can be passed to
                                             :py:attr:`virtual_to_actuated` during the initial
                                             testing of the function. Defaults to None.
        parent (Union(Transformation,KinematicGroup), optional): The transformation or group
                                                                 preceding the
                                                                 :py:class:`KinematicGroup`.
                                                                 Defaults to None.

    Raises:
        ValueError: 'Error: Actuated state is missing.
                     You provided a mapping to actuate the group but no state to be actuated.'
                    if there is no :py:attr:`actuated_state` despite a mapping being passed
        ValueError: 'Error: Only one mapping provided. You need mappings for both ways.
                     Consider to pass a trivial mapping.'
                    if either :py:attr:`actuated_to_virtual` or :py:attr:`virtual_to_actuated`
                    was not set despite providing a :py:attr:`actuated_state`.
        ValueError: 'Error: Mappings missing. You provided an actuated state but no mappings.
                    If you want a trivial mapping you don\'t need to pass an actuated state.
                    Trip will generate one for you.'
                    if both :py:attr:`actuated_to_virtual` and :py:attr:`virtual_to_actuated`
                    were not set despite providing a :py:attr:`actuated_state`.
        RuntimeError: "actuated_to_virtual does not fit virtual state" if the
                      :py:attr:`actuated_to_virtual` function does not return a valid
                      :py:attr:`virtual_state` dictionary.
        RuntimeError: "virtual_to_actuated does not fit actuated state" if the
                      :py:attr:`virtual_to_actuated` function does not return a valid
                      :py:attr:`actuated_state` dictionary.

    """

    @staticmethod
    def object_list_to_key_lists(object_lst):
        """Helper function which transforms dictionary into list of keys.

        Args:
            object_lst (Dict): The dictionary to be transformed

        Returns:
            list(str): A list of keys
        """
        return list(object_lst.keys())

    def __init__(self, name: str, virtual_chain: List[Transformation],
                 actuated_state: Dict[str, float],
                 actuated_to_virtual: Callable,
                 virtual_to_actuated: Callable,
                 act_to_virt_args=None, virt_to_act_args=None,
                 parent=None):

        self._name = name
        self.children = []

        if parent is None:
            self.parent = name
        elif isinstance(parent, (KinematicGroup, Transformation)):
            self.parent = str(parent)
            parent.add_children(name)
        else:
            raise TypeError(
                "The parent of a group must be a either a KinematicGroup or a Transformation")

        self._virtual_chain = {}
        virtual_state = {}
        root = None
        endeffector = None
        for transformation in virtual_chain:
            self._virtual_chain[str(transformation)] = transformation
            if transformation.get_state() != {}:
                virtual_state[str(transformation)] = transformation.get_state()
            if len(transformation.children) >= 2:
                raise ValueError("The Transformation "+str(transformation) +
                                 "contains more than one child." +
                                 " In a virtual chain each transfromation can only have one child")
            if transformation.children == []:
                endeffector = str(transformation)
            if transformation.parent == str(transformation):
                if root is None:
                    root = str(transformation)
                else:
                    raise ValueError("Transformations "+root+" and "+str(transformation) +
                                     " are both connected to the ground." +
                                     " In a virtual chain only one transformation can be" +
                                     " connected to the ground")

        if virtual_state:
            if (actuated_to_virtual or virtual_to_actuated) and not actuated_state:
                raise ValueError('Error: Actuated state is missing.' +
                                 ' You provided a mapping to actuate the group but' +
                                 ' no state to be actuated.')
            if ((actuated_to_virtual and not virtual_to_actuated) or
                    (not actuated_to_virtual and virtual_to_actuated)):
                raise ValueError('Error: Only one mapping provided.' +
                                 ' You need mappings for both ways.' +
                                 ' Consider to pass a trivial mapping.')
            if not (actuated_to_virtual or virtual_to_actuated):
                raise ValueError('Error: Mappings missing.' +
                                 ' You provided an actuated state but no mappings.' +
                                 ' If you want a trivial mapping you can use' +
                                 ' the OpenKinematicGroup.')
        else:
            if actuated_state or (actuated_to_virtual or virtual_to_actuated):
                raise ValueError('Error: Allthough the Group is static, a actuated state and' +
                                 ' or mapping was provided that is not a None object.')

        # build a sorted list of transformation keys from root to endeffector
        # later used in get_transformation_matrix
        current_key = endeffector
        self._chain_keys = [endeffector]
        current_parent = self._virtual_chain[current_key].parent
        while current_parent != current_key:
            next_trafo = self._virtual_chain[current_parent]
            current_key = current_parent
            current_parent = next_trafo.parent
            self._chain_keys.append(current_key)
        self._chain_keys.reverse()

        # check the consistency of the provided mapping functions if the group is not static
        if actuated_state:
            self._original_actuated_to_virtual = actuated_to_virtual
            if act_to_virt_args:
                self._actuated_to_virtual = lambda state: actuated_to_virtual(
                    state, *act_to_virt_args)
            self._actuated_to_virtual = actuated_to_virtual
            actuated_to_virtual_to_check = self._actuated_to_virtual(
                actuated_state)

            if KinematicGroup.object_list_to_key_lists(actuated_to_virtual_to_check) != \
               KinematicGroup.object_list_to_key_lists(virtual_state):
                raise RuntimeError(
                    "actuated_to_virtual does not fit virtual state")
            self._original_virtual_to_actuated = virtual_to_actuated
            if virt_to_act_args:
                self._virtual_to_actuated = lambda state: virtual_to_actuated(
                    state, *virt_to_act_args)
            self._virtual_to_actuated = virtual_to_actuated
            virtual_to_actuated_to_check = virtual_to_actuated(virtual_state)

            if KinematicGroup.object_list_to_key_lists(virtual_to_actuated_to_check) != \
               KinematicGroup.object_list_to_key_lists(actuated_state):
                raise RuntimeError(
                    "virtual_to_actuated does not fit actuated state")

            # Check if inital values fit actuated_to_virtual's
            # and virtual_to_actuated's calculated values.
            for key in actuated_to_virtual_to_check.keys():
                state = actuated_to_virtual_to_check[key]
                init_values_do_not_match = False
                for state_key in state.keys():
                    if state[state_key] != virtual_state[key][state_key]:
                        init_values_do_not_match = True
            if init_values_do_not_match:
                print('Calculated state values do not match given values!' +
                      'Using set_state() before forward_kinematics() or' +
                      ' inverse_kinematics() is recommended.')

        self.virtual_state = deepcopy(virtual_state)
        self.actuated_state = deepcopy(actuated_state)

    def set_virtual_state(self, state: Dict[str, Dict[str, float]]):
        """Sets the :py:attr:`virtual_state` of the Group and
           automatically updates the corresponding :py:attr:`actuated_state`.

        Args:
            state (Dict[str,Dict[str, float]]): A dictionary containing the members of
                                                :py:attr:`virtual_state` that should be set.
                                                The new values need to be valid state
                                                for the state of the joint.

        Raises:
            RuntimeError: if all  :py:class:`Transformation` objects
                          of :py:attr:`_virtual_chain` are static.
            ValueError: if the state to set is not part of keys of :py:attr:`virtual_state`
        """

        if self.actuated_state is None:
            raise RuntimeError(
                "This is a static group! There is no state to be set")

        if all(key in self.virtual_state.keys()for key in state.keys()):
            for key in state.keys():
                self.virtual_state[key] = state[key]
            self._update_chain()
            self.actuated_state = self._virtual_to_actuated(self.virtual_state)
        else:
            raise ValueError('Error: One or more keys are not part of the virtual state. ' +
                             'correct keys are: '+str(self.virtual_state.keys()))

    def set_actuated_state(self, state: Dict[str, float]):
        """Sets the :py:attr:`actuated_state` of the Group and
           automatically updates the corresponding :py:attr:`virtual_state`.

        Args:
            state (Dict[str, float]): A dictionary containing the members
                                      of :py:attr:`actuated_state` that should be set.


        Raises:
            RuntimeError:  if all  :py:class:`Transformation` objects
                           of :py:attr:`_virtual_chain` are static.
            ValueError: if the state to set is not part of keys of :py:attr:`actuated_state`
        """

        if self.actuated_state is None:
            raise RuntimeError(
                "This is a static group! There is no state to be set.")

        if all(key in self.actuated_state.keys()for key in state.keys()):
            for key in state.keys():
                self.actuated_state[key] = state[key]
            new_virtual_state = self._actuated_to_virtual(self.actuated_state)
            for key in new_virtual_state.keys():
                self.virtual_state[key] = new_virtual_state[key]
            self._update_chain()
        else:
            raise ValueError(
                "Error: One or more keys are not part of the actuated state. correct keys are: "
                + str(self.actuated_state.keys()))

    def __str__(self):
        return self._name

    def get_name(self):
        """Returns the :py:attr:`_name` of the :py:class:`KinematicGroup`

        Returns:
            str: the :py:attr:`_name` attribute
        """
        return deepcopy(self._name)

    def get_virtual_state(self):
        """Returns a copy of the :py:attr:`virtual_state`
           attribute of the :py:class:`KinematicGroup` object.

        Returns:
            Dict[str,Dict[str,float]]: a copy of the :py:attr:`virtual_state`
        """
        return deepcopy(self.virtual_state)

    def get_actuated_state(self):
        """Returns a copy of the :py:attr:`actuated_state`
           attribute of the :py:class:`KinematicGroup` object.

        Returns:
            Dict[str,float]: a copy of the :py:attr:`actuated_state`
        """
        if self.actuated_state:
            return deepcopy(self.actuated_state)
        return None

    def get_transformation_matrix(self):
        """Calculates the full transformationmatrix from
           the start of the virtual chain to its endeffector.

        Returns:
            array: The homogenous transformation matrix from
                   the start of the virtual chain to its endeffector.
        """

        # Identity matrix
        transformation = identity_transformation()
        transformations = self._virtual_chain
        for key in self._chain_keys:
            part = transformations[key]
            hmt = part.get_transformation_matrix()
            transformation = transformation @ hmt

        return transformation

    def get_virtual_chain(self):
        """Returns a copy of the :py:attr:`_virtual_chain`
           attribute of a :py:class:`KinematicGroup` object.

        Returns:
            Dict[str,Transformation]: a copy of the :py:attr:`_virtual_chain`
        """
        return deepcopy(self._virtual_chain)

    def _update_chain(self):
        """Propagates changes from the :py:attr:`virtual_state`
           to the underlying :py:class:`Transformation` objects.
        """
        for key in self.virtual_state.keys():
            self._virtual_chain[key].set_state(self.virtual_state[key])

    def add_children(self, child: str):
        """Adds the name of a :py:class:`KinematicGroup` or :py:class:`Transformation`
           as a child.

        Args:
            child (str): the name of a :py:class:`KinematicGroup` or :py:class:`Transformation`
        """
        self.children.append(child)

    def pass_arg_v_to_a(self, argv):
        """Allows arguments to be passed the :py:attr:`virtual_to_actuated` mapping.

        Args:
            argv ([type]): arguments to be passed.
        """
        g_map = self._original_virtual_to_actuated
        self._virtual_to_actuated = lambda state: g_map(state, *argv)

    def pass_arg_a_to_v(self, argv):
        """Allows arguments to be passed the :py:attr:`actuated_to_virtual` mapping.

        Args:
            argv ([type]): arguments to be passed.
        """
        f_map = self._original_actuated_to_virtual
        self._actuated_to_virtual = lambda state: f_map(state, *argv)


class OpenKinematicGroup(KinematicGroup):
    """A subclass of the :py:class:`KinematicGroup` that assumes that all states
       of the virtual_chain are actuated and automatically generates mappings.
       Typically only used internally by the :py:class`Robot` class to convert
       :py:class`Transformation` objects to :py:class`KinematicGroup`s.

    Args:
        name (str): The unique name identifying the group.
                    No two :py:class:`KinematicGroup` objects of a :py:class`Robot` should have the
                    same name
        virtual_chain (List[Transformation]): A list of :py:class:`Transformation`
                                              objects forming a serial Kinematic chain.
        parent (Union(Transformation,KinematicGroup), optional): The transformation or group
                                                                 preceding the
                                                                 :py:class:`KinematicGroup`.
                                                                 Defaults to None.

    """

    def __init__(self, name: str, virtual_chain: List[Transformation], parent=None):
        virtual_state = {}
        virtual_trafo_dict = {}
        for transformation in virtual_chain:
            virtual_trafo_dict[str(transformation)] = transformation
            if transformation.get_state() != {}:
                virtual_state[str(transformation)] = transformation.get_state()

        if any(virtual_state):  # group is dynamic
            actuated_state_dummy = {}
            f_map = {}
            g_map = {}

            for virtual_key, transformation in virtual_trafo_dict.items():

                state = transformation.get_state()
                for key, value in state.items():
                    concat_key = transformation.get_name() + '_' + key
                    actuated_state_dummy.setdefault(concat_key, value)
                    f_map.setdefault(concat_key, (key, virtual_key))
                    g_map.setdefault((key, virtual_key), concat_key)

            # create trivial mappings
            def trivial_actuated_to_virtual(state):
                out = {}
                for concat_key, value in state.items():
                    key, index = f_map[concat_key]
                    out[index] = {key: value}
                return out

            def trivial_virtual_to_actuated(states):
                out = {}
                for virtual_key in states.keys():
                    transformation = states[virtual_key]
                    for key in transformation.keys():
                        combined_key = (key, virtual_key)
                        new_key = g_map[combined_key]
                        out[new_key] = states[virtual_key][key]
                return out

            actuated_to_virtual = trivial_actuated_to_virtual
            virtual_to_actuated = trivial_virtual_to_actuated
            actuated_state = actuated_state_dummy
        else:   # group is static
            actuated_state = None
            actuated_to_virtual = None
            virtual_to_actuated = None
        super(OpenKinematicGroup, self).__init__(name=name,
                                                 virtual_chain=virtual_chain,
                                                 actuated_state=actuated_state,
                                                 actuated_to_virtual=actuated_to_virtual,
                                                 virtual_to_actuated=virtual_to_actuated,
                                                 parent=parent)



from numpy import array
from casadi import cos, sin


def identity_transformation():
    """Returns a 4x4 identity matix

    Returns:
        numy.array: a 4x4 identity matrix
    """
    return array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]], dtype=object)


def hom_translation_matrix(t_x=0, t_y=0, t_z=0):
    """Returns a homogenous translation matrix

    Args:
        t_x (int, optional): Translation along the x axis. Defaults to 0.
        t_y (int, optional): Translation along the y axis. Defaults to 0.
        t_z (int, optional): Translation along the z axis. Defaults to 0.

    Returns:
        numpy.array: A 4x4 homogenous translation matrix
    """
    return array(
        [[1, 0, 0, t_x], [0, 1, 0, t_y], [0, 0, 1, t_z], [0., 0., 0., 1.]], dtype=object)


def hom_rotation(rotation_matrix):
    """Converts a 3x3 rotation matrix into a 4x4 homogenous rotation mtrix

    Args:
        rotation_matrix (numpy.array): A 3x3 rotation matrix

    Returns:
        numpy.array:  A 4x4 homogenous rotation matrix
    """
    matrix = identity_transformation()
    matrix[:3, :3] = rotation_matrix
    return matrix


def quat_rotation_matrix(q_w, q_x, q_y, q_z) -> array:
    """Generates a 3x3 rotation matrix from q quaternion

    Args:
        q_w (float): part of a quaternion [q_w,q_x,q_y,q_z]
        q_x (float): part of a quaternion [q_w,q_x,q_y,q_z]
        q_y (float): part of a quaternion [q_w,q_x,q_y,q_z]
        q_z (float): part of a quaternion [q_w,q_x,q_y,q_z]

    Returns:
        numpy.array: A 3x3 rotation matrix
    """
    return array([[1-2*(q_y**2+q_z**2), 2*(q_x*q_y-q_z*q_w), 2*(q_x*q_z + q_y*q_w)],
                 [2*(q_x*q_y + q_z*q_w), 1-2 *
                     (q_x**2+q_z**2), 2*(q_y*q_z - q_x*q_w)],
                 [2*(q_x*q_z-q_y*q_w), 2*(q_y*q_z+q_x*q_w), 1-2*(q_x**2+q_y**2)]], dtype=object)


def x_axis_rotation_matrix(theta):
    """Generates a matrix rotating around the x axis

    Args:
        theta (float): The angle of rotation in rad

    Returns:
        numpy.array: A 3x3 rotation matrix
    """
    return array([[1, 0, 0],
                  [0, cos(theta), -sin(theta)],
                  [0, sin(theta), cos(theta)]], dtype=object)


def y_axis_rotation_matrix(theta):
    """Generates a matrix rotating around the y axis

    Args:
        theta (float): The angle of rotation in rad

    Returns:
        numpy.array: A 3x3 rotation matrix
    """
    return array([[cos(theta), 0, sin(theta)],
                  [0, 1, 0],
                  [-sin(theta), 0, cos(theta)]], dtype=object)


def z_axis_rotation_matrix(theta):
    """Generates a matrix rotating around the z axis

    Args:
        theta (float): The angle of rotation in rad

    Returns:
        numpy.array: A 3x3 rotation matrix
    """
    return array([[cos(theta), -sin(theta), 0],
                  [sin(theta), cos(theta), 0],
                  [0, 0, 1]], dtype=object)


def get_rotation(matrix):
    """Returns the 3x3 rotation matrix of the :py:class`TransformationMatrix`

    Returns:
        numpy.array: The 3x3 rotation matrix
    """
    return matrix[: 3, : 3]


def get_translation(matrix):
    """Returns the translation of the :py:class`TransformationMatrix`

    Returns:
        numpy array: The 3 dimensional translation
    """
    return matrix[: 3, 3]



from typing import Dict, List
from copy import deepcopy

from trip_kinematics.Utility import hom_translation_matrix, x_axis_rotation_matrix
from trip_kinematics.Utility import y_axis_rotation_matrix, z_axis_rotation_matrix
from trip_kinematics.Utility import quat_rotation_matrix
import trip_kinematics.KinematicGroup


def array_find(arr, obj) -> int:
    """A helpher function which finds the index of an object in an array.
       Instead of throwing an error when no index can be found it returns -1.

    Args:
        arr : the array to be searched
        obj : the object whose index is to be found.

    Returns:
        int: The index of obj in the array. -1 if the object is not in the array
    """
    index = -1
    try:
        index = arr.index(obj)
        return index
    except:
        return -1


class Transformation():
    """Initializes the :py:class:`Transformation` class.

    Args:
        name (str): The unique name identifying the Transformation.
                    No two :py:class:`Transformation` objects of a :py:class`Robot`
                    should have the same name
        values (Dict[str, float]): A parametric description of the transformation.
        state_variables (List[str], optional): This list describes which state variables are
                                               dynamically changable.
                                               This is the case if the :py:class:`Transformation`
                                               represents a joint. Defaults to [].

    Raises:
        ValueError: A dynamic state was declared that does not
                    correspond to a parameter declared in ``values``.
    """

    @staticmethod
    def get_convention(state: Dict[str, float]):
        """Returns the connvention which describes how the matrix
           of a :py:class:`Transformation` is build from its state.

        Args:
            state (Dict[str, float]): :py:attr:'state'

        Raises:
            ValueError: "Invalid key." If the dictionary kontains keys that dont
                        correspond to a parameter of the transformation.
            ValueError: "State can't have euler angles and quaternions!"
                        If the dictionary contains keys correspondig
                        to multiple mutually exclusive conventions.

        Returns:
            [type]: A string describing the convention
        """

        valid_keys = ['tx', 'ty', 'tz', 'qw', 'qx',
                      'qy', 'qz', 'rx', 'rz', 'ry']

        quaternion_keys = valid_keys[3:7]
        euler_keys = valid_keys[7:]
        got_quaternion = False
        got_euler = False

        for key in state.keys():
            if array_find(valid_keys, key) == -1:
                raise ValueError(
                    "Invalid key, acceptable keys are: "+str(valid_keys))
            if array_find(quaternion_keys, key) >= 0:
                got_quaternion = True
            if array_find(euler_keys, key) >= 0:
                got_euler = True

        if got_euler and got_quaternion:
            raise ValueError("State can't have euler angles and quaternions!")

        if got_euler:
            return "euler"
        return "quaternion"

    def __init__(self, name: str, values: Dict[str, float],
                 state_variables: List[str] = None, parent=None):

        self._name = name
        self.children = []

        if parent is None:
            self.parent = name
        elif isinstance(parent, (trip_kinematics.KinematicGroup, Transformation)):
            self.parent = str(parent)
            parent.add_children(name)
        else:
            raise TypeError(
                "The parent of a Transformation must be either " +
                "a KinematicGroup or a Transformation")

        if state_variables is None:
            state_variables = []
        if not set(state_variables) <= set(values.keys()):
            raise ValueError(
                "Element of state_variables do not correspond to key(s) of values dictionary." +
                " This can happen if both follow different conventions")

        constants = {}
        state = {}

        self.convention = Transformation.get_convention(values)

        for key in values.keys():
            if array_find(state_variables, key) != -1:
                state.setdefault(key, values.get(key))
            else:
                constants.setdefault(key, values.get(key))

        self._state = state
        self._constants = constants

    def __str__(self):
        return self._name

    def get_state(self):
        """Returns a copy of the :py:attr:`_state`
           attribute of the :py:class:`Transformation` object.

        Returns:
            Dict[str,float]: a copy of the :py:attr:`_state`
        """
        return deepcopy(self._state)

    def set_state(self, state: Dict[str, float]):
        """Sets the state of the :py:class:`Transformation` object.


        Args:
            state (Dict[str, float]): Dictionary with states that should be set.
                                      Does not have to be the full state.

        Raises:
            KeyError: If a key in the argument is not valid state parameter name.
        """
        for key in state.keys():
            if not key in self._state.keys():
                raise KeyError(
                    "The specified keys is not part of the Transformations state." +
                    "Maybe the Transofmration uses a different convention?")
            self._state[key] = state[key]

    def get_name(self):
        """Returns the :py:attr:`_name` of the :py:class:`Transformation`

        Returns:
            str: a copy the :py:attr:`_name` attribute
        """
        return deepcopy(self._name)

    def get_transformation_matrix(self):
        """Returns a homogeneous transformation matrix build
           from the :py:attr:`state` and :py:attr:`constants`

        Raises:
            RuntimeError: If the convention used in :py:attr:`state` is not supported.
                          Should normally be catched during initialization.

        Returns:
            [type]: A transformation matrix build using the parameters
                    of the :py:class:`Transformation` :py:attr:`state`
        """

        # collect transformation parameters from state and constants respectively
        t_x = self._constants.get(
            'tx') if 'tx' in self._constants.keys() else self._state.get('tx', 0)
        t_y = self._constants.get(
            'ty') if 'ty' in self._constants.keys() else self._state.get('ty', 0)
        t_z = self._constants.get(
            'tz') if 'tz' in self._constants.keys() else self._state.get('tz', 0)

        matrix = hom_translation_matrix(t_x, t_y, t_z)

        if self.convention == 'euler':
            r_x = self._constants.get(
                'rx') if 'rx' in self._constants.keys() else self._state.get('rx', 0)
            r_y = self._constants.get(
                'ry') if 'ry' in self._constants.keys() else self._state.get('ry', 0)
            r_z = self._constants.get(
                'rz') if 'rz' in self._constants.keys() else self._state.get('rz', 0)

            matrix[:3, :3] = x_axis_rotation_matrix(
                r_x) @ y_axis_rotation_matrix(r_y) @ z_axis_rotation_matrix(r_z)

        elif self.convention == 'quaternion':
            q_w = self._constants.get(
                'qw') if 'qw' in self._constants.keys() else self._state.get('qw', 0)
            q_x = self._constants.get(
                'qx') if 'qx' in self._constants.keys() else self._state.get('qx', 0)
            q_y = self._constants.get(
                'qy') if 'qy' in self._constants.keys() else self._state.get('qy', 0)
            q_z = self._constants.get(
                'qz') if 'qz' in self._constants.keys() else self._state.get('qz', 0)

            matrix[:3, :3] = quat_rotation_matrix(q_w, q_x, q_y, q_z)
        else:
            raise RuntimeError("No Convention. " +
                               "This should normally be catched during initialization. " +
                               "Did you retroactively change the keys of the Transformation state?")

        return matrix

    def add_children(self, child: str):
        """Adds the name of a :py:class:`KinematicGroup` or :py:class:`Transformation`
           as a child.

        Args:
            child (str): the name of a :py:class:`KinematicGroup` or :py:class:`Transformation`
        """
        self.children.append(child)



from typing import Dict
from math import radians
import casadi
import numpy as np

from trip_kinematics.Utility import hom_rotation, get_translation
from trip_kinematics.Utility import hom_translation_matrix, y_axis_rotation_matrix
from trip_kinematics.KinematicGroup import KinematicGroup
from trip_kinematics.Transformation import Transformation
from trip_kinematics.Robot import Robot


l_1 = 1
l_2 = 0.7
l_3 = 1.2
l_4 = 0.4
l_5 = 1.7

# zero conventions for the actuated joints
a_1_offset = 0
a_2_offset = 0

virtual_joint_1 = Transformation(name="q_1",
                                 values={'ry': 0},
                                 state_variables=['ry'])
link_1 = Transformation(name="link_1",
                        values={'tx': l_1+l_3+0.4},
                        parent=virtual_joint_1)
virtual_joint_2 = Transformation(name="q_2",
                                 values={'ry': radians(-90)},
                                 state_variables=['ry'])
link_2 = Transformation(name="link_2",
                        values={'tx': l_5},
                        parent=virtual_joint_2)


################################################
# Direct mappings using geometric calculations #
################################################

def geometric_q_to_a_group_1(state: Dict[str, float], tips: Dict[str, float] = None):
    # convert joint angle to triangle angle
    q_1 = radians(90) - state['q_1']['ry']
    return {'a_1': np.sqrt(l_1**2+l_2**2-2*l_1*l_2*np.cos(q_1))}


def geometric_a_to_q_group_1(state: Dict[str, float], tips: Dict[str, float] = None):
    a_1 = state['a_1'] + a_1_offset
    return {'q_1': {'ry': np.arccos((l_1**2+l_2**2-a_1**2)/(2*l_1*l_2))}}


def geometric_q_to_a_group_2(state: Dict[str, float], tips: Dict[str, float] = None):
    q_2 = -1 * state['q_2']['ry']  # convert joint angle to triangle angle
    return {'a_2': np.sqrt(l_3**2+l_4**2-2*l_3*l_4*np.cos(q_2))}


def geometric_a_to_q_group_2(state: Dict[str, float], tips: Dict[str, float] = None):
    a_2 = state['a_2'] + a_2_offset
    return {'q_2': {'ry': np.arccos((l_3**2+l_4**2-a_2**2)/(2*l_3*l_4))}}


geometric_group_1 = KinematicGroup(name="geometric group 1",
                                   virtual_chain=[virtual_joint_1, link_1],
                                   actuated_state={'a_1': 0},
                                   actuated_to_virtual=geometric_a_to_q_group_1,
                                   virtual_to_actuated=geometric_q_to_a_group_1)

geometric_group_2 = KinematicGroup(name="geometric group 2",
                                   virtual_chain=[virtual_joint_2, link_2],
                                   actuated_state={'a_2': 0},
                                   actuated_to_virtual=geometric_a_to_q_group_2,
                                   virtual_to_actuated=geometric_q_to_a_group_2,
                                   parent=geometric_group_1)

geometric_excavator = Robot([geometric_group_1, geometric_group_2])


###################################################################
# mappings using colsure equation solution geometric calculations #
###################################################################
opts = {'ipopt.print_level': 0, 'print_time': 0}

closure_1_state = casadi.SX.sym('cls_1_q', 3)

cls_q_1 = hom_rotation(y_axis_rotation_matrix(closure_1_state[0]))
cls_l_1 = hom_translation_matrix(t_x=l_1)
cls_qs_2 = hom_rotation(y_axis_rotation_matrix(closure_1_state[1]))
cls_a_1 = hom_translation_matrix(t_x=closure_1_state[2])
cls_a_1z = hom_translation_matrix(t_x=a_1_offset)
cls_1_trafo = cls_q_1 * cls_l_1 * cls_qs_2 * cls_a_1 * cls_a_1z

cls_1_trafo_pos = get_translation(cls_1_trafo)

c_1 = (cls_1_trafo_pos[0]-l_2)**2 + \
    cls_1_trafo_pos[1]**2 + cls_1_trafo_pos[2]**2


def closure_q_to_a_group_1(state: Dict[str, float]):
    nlp = {'x': closure_1_state[1:], 'f': c_1, 'p': closure_1_state[0]}
    nlp_solver = casadi.nlpsol('q_to_a', 'ipopt', nlp, opts)
    solution = nlp_solver(x0=[0, 0], p=[state['q_1']['ry']])
    sol_vector = np.array(solution['x'])
    return {'a_1': sol_vector[1]}


def closure_a_to_q_group_1(state: Dict[str, float]):
    nlp = {'x': closure_1_state[:1], 'f': c_1, 'p': closure_1_state[2]}
    nlp_solver = casadi.nlpsol('a_to_q', 'ipopt', nlp, opts)
    solution = nlp_solver(x0=[0, 0], p=[state['a_1']])
    sol_vector = np.array(solution['x'])
    return {'q_1': {'ry': sol_vector[0]}}


closure_2_state = casadi.SX.sym('cls_2_q', 3)

cls_q_2 = hom_rotation(y_axis_rotation_matrix(closure_2_state[0]))
cls_l_4 = hom_translation_matrix(t_x=l_4)
cls_qs_4 = hom_rotation(y_axis_rotation_matrix(closure_2_state[1]))
cls_a_2 = hom_translation_matrix(t_x=closure_2_state[2])
cls_a_2z = hom_translation_matrix(t_x=a_1_offset)
cls_2_trafo = cls_q_2 * cls_l_4 * cls_qs_4 * cls_a_2 * cls_a_2z

cls_2_trafo_pos = get_translation(cls_1_trafo)

c_2 = (cls_2_trafo_pos[0]+l_3)**2 + \
    cls_2_trafo_pos[1]**2 + cls_2_trafo_pos[2]**2


def closure_q_to_a_group_1(state: Dict[str, float]):
    nlp = {'x': closure_2_state[1:], 'f': c_2, 'p': closure_2_state[0]}
    nlp_solver = casadi.nlpsol('q_to_a', 'ipopt', nlp, opts)
    solution = nlp_solver(x0=[0, 0], p=[state['q_1']['ry']])
    sol_vector = np.array(solution['x'])
    return {'a_2': sol_vector[1]}


def closure_a_to_q_group_1(state: Dict[str, float]):
    nlp = {'x': closure_2_state[:1], 'f': c_2, 'p': closure_2_state[2]}
    nlp_solver = casadi.nlpsol('a_to_q', 'ipopt', nlp, opts)
    solution = nlp_solver(x0=[0, 0], p=[state['a_1']])
    sol_vector = np.array(solution['x'])
    return {'q_2': {'ry': sol_vector[0]}}


closure_group_1 = KinematicGroup(name="geometric group 1",
                                 virtual_chain=[virtual_joint_1, link_1],
                                 actuated_state={'a_1': 0},
                                 actuated_to_virtual=geometric_a_to_q_group_1,
                                 virtual_to_actuated=geometric_q_to_a_group_1)

closure_group_2 = KinematicGroup(name="geometric group 2",
                                 virtual_chain=[virtual_joint_2, link_2],
                                 actuated_state={'a_2': 0},
                                 actuated_to_virtual=geometric_a_to_q_group_2,
                                 virtual_to_actuated=geometric_q_to_a_group_2,
                                 parent=geometric_group_1)

closure_excavator = Robot([closure_group_1, closure_group_2])



from typing import Dict, List
from copy import deepcopy
from casadi import SX

from trip_kinematics.Utility import identity_transformation
from trip_kinematics.KinematicGroup import KinematicGroup, OpenKinematicGroup
from trip_kinematics.Transformation import Transformation


class Robot:
    """A class representing the kinematic model of a robot.

    Args:
        kinematic_chain (List[KinematicGroup]): A list of Kinematic Groups and Transformations
                                                which make up the robot.
                                                Transformations are automatically
                                                converted to groups

    Raises:
        KeyError: "More than one robot actuator has the same name!
                   Please give each actuator a unique name"
                  if there are actuated states with the same names between the
                  :py:class`KinematicGroup` objects of the :py:class`Robot`
        KeyError: if there are joints with the same names between
                  the :py:class`KinematicGroup` objects of the :py:class`Robot`
    """

    def __init__(self, kinematic_chain: List[KinematicGroup]) -> None:

        self._group_dict = {}
        self._actuator_group_mapping = {}
        self._virtual_group_mapping = {}
        for i, group in enumerate(kinematic_chain):
            if isinstance(group, Transformation):
                # create group from trafo, lifting the parent and child relations to group level
                group_children = group.children
                group_parent = group.parent

                # KinematicGroup expects endeffector of virtual children to have no children
                group.children = []
                # KinematicGroup expects root of virtual children to have itself as a parent
                group.parent = str(group)

                group = OpenKinematicGroup(name=str(group), virtual_chain=[group],
                                           parent=self._group_dict[str(kinematic_chain[i-1])])

                # carefull this circumvents the usual checks for correct group and parent names!
                group.children = group_children
                group.parent = group_parent

            self._group_dict[str(group)] = group
            if group.get_virtual_state() != {}:
                group_actuators = group.get_actuated_state().keys()
                for key in group_actuators:
                    if key in self._actuator_group_mapping.keys():
                        raise KeyError(
                            "More than one robot actuator has the same name!" +
                            " Please give each actuator a unique name")
                    self._actuator_group_mapping[key] = str(group)

                for key in group.get_virtual_state().keys():
                    if key in self._virtual_group_mapping.keys():
                        raise KeyError(
                            "More than one transformation of a virtual chain has the same name!" +
                            " Please give each virtual transformation of a robot a unique name")
                    self._virtual_group_mapping[key] = str(group)

    def get_groups(self):
        """Returns a dictionary of the py:class`KinematicGroup` managed by the :py:class`Robot`_
           Since Transformations are internally converted to Groups, this also returns all
           Transformations.

        Returns:
            Dict[str, KinematicGroup]: The dictionary of py:class`KinematicGroup` objects.
        """
        return deepcopy(self._group_dict)

    def pass_group_arg_v_to_a(self, argv_dict: Dict):
        """Passes optional virtual_to_actuated mapping arguments
           to :py:class`KinematicGroup` objects of the robot.

        Args:
            argv_dict (Dict): A dictionary containing the mapping arguments keyed with the
                              :py:class`KinematicGroup` names.

        Raises:
            KeyError: If no group with the name given in the argument is part of the robot.
        """
        for key in argv_dict.keys():
            if key not in self._group_dict.keys():
                raise KeyError("No group with name "+str(key)+"in this robot")
            self._group_dict[key].pass_arg_v_to_a(argv_dict[key])

    def pass_group_arg_a_to_v(self, argv_dict):
        """Passes optional actuated_to_virtual mapping arguments
           to :py:class`KinematicGroup` objects of the robot.

        Args:
            argv_dict (Dict): A dictionary containing the mapping arguments keyed with the
                              :py:class`KinematicGroup` names.

        Raises:
            KeyError: If no group with the name given in the argument is part of the robot.
        """
        for key in argv_dict.keys():
            if key not in self._group_dict.keys():
                raise KeyError("No group with name "+str(key)+"in this robot")
            self._group_dict[key].pass_arg_a_to_v(argv_dict[key])

    def set_virtual_state(self, state: Dict[str, Dict[str, float]]):
        """Sets the virtual state of multiple virtual joints of the robot.

        Args:
            state (Dict[str,Dict[str, float]]): A dictionary containing the members of
                                                 :py:attr:`__virtual_state` that should be set.
                                                The new values need to be valid state
                                                for the state of the joint.
        """
        for key in state.keys():
            virtual_state = {key: state[key]}
            self._group_dict[self._virtual_group_mapping[key]
                             ].set_virtual_state(virtual_state)

    def set_actuated_state(self, state: Dict[str, float]):
        """Sets the virtual state of multiple actuated joints of the robot.

        Args:
            state (Dict[str, float]):  A dictionary containing the members of
                                        :py:attr:`__actuated_state` that should be set.
        """
        grouping = {}
        for key in state.keys():
            if self._actuator_group_mapping[key] not in grouping.keys():
                grouping[self._actuator_group_mapping[key]] = {}
            grouping[self._actuator_group_mapping[key]][key] = state[key]
        for key, group_state in grouping.items():
            self._group_dict[key].set_actuated_state(group_state)

    def get_actuated_state(self):
        """Returns the actuated state of the :py:class`Robot` comprised
           of the actuated states of the individual :py:class`KinematicGroup`.

        Returns:
            Dict[str, float]: combined actuated state of all :py:class`KinematicGroup` objects.
        """
        actuated_state = {}
        for group in self._group_dict.values():
            actuated_group_state = group.get_actuated_state()
            if actuated_group_state is not None:
                for actuated_key in actuated_group_state:
                    actuated_state[actuated_key] = actuated_group_state[actuated_key]
        return actuated_state

    def get_virtual_state(self):
        """Returns the virtual state of the :py:class`Robot` comprised
           of the virtual states of the individual :py:class`KinematicGroup`.

        Returns:
            Dict[str,Dict[str, float]]: combined virtual state of all
                                         :py:class`KinematicGroup` objects.
        """
        virtual_state = {}
        for group_key in self._group_dict.keys():
            group_state = self._group_dict[group_key].get_virtual_state()
            if group_state != {}:
                for key in group_state.keys():
                    virtual_state[key] = group_state[key]
        return virtual_state

    def get_symbolic_rep(self, endeffector: str):
        """This Function returnes a symbolic representation of the virtual chain.

        Args:
            endeffector (str): The name of the group whose virtual chain
                               models the desired endeffector

        Raises:
            KeyError: If the endeffector argument is not the name of a transformation or group

        Returns:
            SX: A 4x4 symbolic casadi matrix containing the transformation from base to endeffector
        """

        matrix = identity_transformation()

        symbolic_state = []
        symbolic_keys = []

        group_dict = self.get_groups()
        if endeffector not in group_dict.keys():
            raise KeyError(
                "The endeffector must be a valid group or transformation name." +
                " Valid names for this robot are: "+str(group_dict.keys()))

        endeff_group = group_dict[endeffector]
        current_parent = endeff_group.parent
        current_key = endeffector
        group_key_list = [endeffector]

        while current_parent != current_key:
            next_group = group_dict[current_parent]
            current_key = current_parent
            current_parent = next_group.parent
            group_key_list.append(current_key)

        group_key_list.reverse()
        for group_key in group_key_list:
            group = group_dict[group_key]
            virtual_trafo = group.get_virtual_chain()

            for virtual_key in virtual_trafo.keys():
                virtual_transformation = virtual_trafo[virtual_key]
                state = virtual_transformation.get_state()

                if state != {}:
                    for key in state.keys():
                        state[key] = SX.sym(virtual_key+'_'+key)
                        symbolic_state.append(state[key])
                        symbolic_keys.append([virtual_key, key])
                symbolic_transformation = deepcopy(virtual_transformation)
                symbolic_transformation.set_state(state)

                hmt = symbolic_transformation.get_transformation_matrix()
                matrix = matrix @ hmt

        hom_matrix = SX.zeros(4, 4)
        for i in range(4):
            for j in range(4):
                hom_matrix[i, j] = matrix[i, j]

        return hom_matrix, symbolic_state, symbolic_keys

    def get_endeffectors(self):
        """Returns a list of possible endeffectors.
           These are the names of all :py:class:`KinematicGroup` objects.
           Since Transformations are internally converted to Groups,
           this includes the names of all Transformations.

        Returns:
            list(str): list of possible endeffectors.
        """
        return list(self.get_groups().keys())


def forward_kinematics(robot: Robot, endeffector):
    """Calculates a robots transformation from base to endeffector using its current state

    Args:
        robot (Robot): The robot for which the forward kinematics should be computed

    Returns:
        numpy.array : The Transformation from base to endeffector
    """
    transformation = identity_transformation()
    group_dict = robot.get_groups()
    if endeffector not in group_dict.keys():
        raise KeyError(
            "The endeffector must be a valid group name. Valid group names for this robot are: " +
            str(group_dict.keys()))
    endeff_group = group_dict[endeffector]
    current_parent = endeff_group.parent
    current_key = endeffector
    group_key_list = [endeffector]
    while current_parent != current_key:
        next_group = group_dict[current_parent]
        current_key = current_parent
        current_parent = next_group.parent
        group_key_list.append(current_key)

    group_key_list.reverse()
    for group_key in group_key_list:
        group = group_dict[group_key]
        hmt = group.get_transformation_matrix()
        transformation = transformation @ hmt
    return transformation



from copy import deepcopy
from typing import Dict
from math import radians

from trip_kinematics.KinematicGroup import KinematicGroup, Transformation
from trip_kinematics.Robot import Robot
from trip_robots.triped_leg import swing_to_gimbal, gimbal_to_swing


def leg_model(leg_number: str):
    """Helper function that constructs each TriPed leg as a list of Transformations.

    Args:
        leg_number (str): The leg number which determins the orientation.
                          Acceptable values [0,1,2]
    """
    def rename_swing_to_gimbal(swing: Dict[str, float], tips: Dict[str, float] = None):
        swing = deepcopy(swing)
        swing['swing_left'] = swing[leg_name+'swing_left']
        swing['swing_right'] = swing[leg_name+'swing_right']
        del swing[leg_name+'swing_left']
        del swing[leg_name+'swing_right']

        gimbal = swing_to_gimbal(swing, tips)

        gimbal[leg_name+'gimbal_joint'] = gimbal['gimbal_joint']
        del gimbal['gimbal_joint']

        return gimbal

    def rename_gimbal_to_swing(gimbal: Dict[str, float], tips: Dict[str, float] = None):
        gimbal = deepcopy(gimbal)
        gimbal['gimbal_joint'] = gimbal[leg_name+'gimbal_joint']
        del gimbal[leg_name+'gimbal_joint']

        swing = gimbal_to_swing(gimbal, tips)

        swing[leg_name+'swing_left'] = swing['swing_left']
        swing[leg_name+'swing_right'] = swing['swing_right']
        del swing['swing_left']
        del swing['swing_right']

        return swing

    leg_name = 'leg'+str(leg_number)+'_'

    leg_rotation = Transformation(name=leg_name+'leg_rotation',
                                  values={'rz': -1*radians(120)*leg_number})
    a_ccs_p_trans = Transformation(name=leg_name+'A_ccs_P_trans',
                                   values={'tx': 0.265, 'tz': 0.014},
                                   parent=leg_rotation)
    a_ccs_p_rot = Transformation(name=leg_name+'gimbal_joint',
                                 values={'rx': 0, 'ry': 0, 'rz': 0},
                                 state_variables=['rx', 'ry', 'rz'],
                                 parent=a_ccs_p_trans)

    closed_chain = KinematicGroup(name=leg_name+'closed_chain',
                                  virtual_chain=[leg_rotation,
                                                 a_ccs_p_trans, a_ccs_p_rot],
                                  actuated_state={
                                      leg_name+'swing_left': 0, leg_name+'swing_right': 0},
                                  actuated_to_virtual=rename_swing_to_gimbal,
                                  virtual_to_actuated=rename_gimbal_to_swing)

    a_p_ll = Transformation(name=leg_name+'A_P_LL',
                            values={'tx': 1.640, 'tz': -0.037, },
                            parent=closed_chain)
    a_ll_zero = Transformation(name=leg_name+'zero_angle_convention',
                               values={'ry': radians(-3)},
                               parent=a_p_ll)
    a_ll_zero_ll_joint = Transformation(name=leg_name+'extend_joint',
                                        values={'ry': 0},
                                        state_variables=['ry'],
                                        parent=a_ll_zero)
    a_ll_joint_fcs = Transformation(name=leg_name+'A_LL_Joint_FCS',
                                    values={'tx': -1.5},
                                    parent=a_ll_zero_ll_joint)

    return [closed_chain, a_ll_zero_ll_joint, a_ll_joint_fcs, a_p_ll, a_ll_zero]


triped = Robot(leg_model(0)+leg_model(1)+leg_model(2))
triped.set_actuated_state({'leg0_swing_left': 0, 'leg0_swing_right': 0})



# setup.py
import setuptools

setuptools.setup(
    name='trip_kinematics',
    version='1.0',
    author='Torben Miller, Jan Baumgärtner',
    license='MIT',
    description='...',
    install_requires=['casadi>=3.5.5', 'numpy>=1.17.4, < 1.20.0'],
    packages=['trip_kinematics', 'trip_robots']
)



from typing import Dict
from copy import deepcopy
from casadi import SX, nlpsol, vertcat
from numpy import array

from trip_kinematics.Robot import Robot


class SimpleInvKinSolver:
    """A Simple Kinematic Solver Class that calculates
       the inverse kinematics for a given endeffector.

    Args:
        robot (Robot): The Robot for which the kinematics should be calculated
        endeffector (str): the name of the endeffector
        orientation (bool, optional): Boolean flag deciding if inverse kinematics
                                      targets also specify orientation.
                                      Defaults to False.
        update_robot (bool, optional): Boolean flag decding if the inverse kinematics should
                                       immediately update the robot model.
                                       Defaults to False.
    """

    def __init__(self, robot: Robot, endeffector: str, orientation=False, update_robot=False):

        matrix, symboles, self._symbolic_keys = robot.get_symbolic_rep(
            endeffector)
        self.endeffector = endeffector
        if update_robot:
            self._robot = robot
        else:
            self._robot = deepcopy(robot)

        if orientation is False:
            end_effector_position = SX.sym("end_effector_pos", 3)
            objective = ((matrix[0, 3] - end_effector_position[0])**2 +
                         (matrix[1, 3] - end_effector_position[1])**2 +
                         (matrix[2, 3] - end_effector_position[2])**2)

            nlp = {'x': vertcat(*symboles), 'f': objective,
                   'p': end_effector_position}
            opts = {'ipopt.print_level': 0, 'print_time': 0}
            self.inv_kin_solver = nlpsol('inv_kin', 'ipopt', nlp, opts)

    def solve_virtual(self, target: array, initial_tip=None):
        """Returns the virtual state needed for the endeffector to be in the target position
        Args:
            target (numpy.array): The target state of the endeffector.
                                  Either a 3 dimensional position or a 4x4 homogenous transformation
            initial_tip ([type], optional): Initial state of the solver.
                                            Defaults to None in which case zeros are used.

        Returns:
            Dict(str,Dict(str,float)): The actuated state leading the endeffector
                                       to the target position.
        """
        if initial_tip is None:
            initial_guess = [0]*len(self._symbolic_keys)
        else:
            initial_guess = self._virtual_to_solver_state(initial_tip)
        if len(initial_guess) != len(self._symbolic_keys):
            raise RuntimeError("The initial state has "+str(len(initial_guess)) +
                               " values, while the solver expected ", str(len(self._symbolic_keys)))
        solution = self.inv_kin_solver(x0=initial_guess, p=target)
        return self._solver_to_virtual_state(solution['x'])

    def solve_actuated(self, target: array, initial_tip=None, mapping_argument=None):
        """Returns the actuated state needed for the endeffector to be in the target position
        Args:
            target (numpy.array): The target state of the endeffector.
                                  Either a 3 dimensional position or a 4x4 homogenous transformation
            initial_tip ([type], optional): Initial state of the solver.
                                            Defaults to None in which case zeros are used.
            mapping_argument ([type], optional): Optional arguments for the virtual_to_actuated
                                                 mappings of the robot.
                                                 Defaults to None.

        Returns:
            Dict(str,float): The actuated state leading the endeffector to the target position.
        """
        virtual_state = self.solve_virtual(
            target=target, initial_tip=initial_tip)
        if mapping_argument is not None:
            self._robot.pass_group_arg_v_to_a(mapping_argument)

        self._robot.set_virtual_state(virtual_state)
        actuated_state = self._robot.get_actuated_state()
        return actuated_state

    def _solver_to_virtual_state(self, solver_state):
        """This function maps the solution of a casadi solver to the virtual state of a robot.

        Args:
            solver_state ([type]): A solution of a nlp solver
        Returns:
            Dict[str,Dict[str, float]]: a :py:attr:`virtual_state` of a robot.
        """
        virtual_state = {}
        # convert casadi DM to usable datatype
        solver_state = array(solver_state)
        for i, solver_state_value in enumerate(solver_state):
            outer_key = self._symbolic_keys[i][0]
            inner_key = self._symbolic_keys[i][1]
            if outer_key not in virtual_state.keys():
                virtual_state[outer_key] = {}

            virtual_state[outer_key][inner_key] = solver_state_value[0]
        return virtual_state

    def _virtual_to_solver_state(self, virtual_state: Dict[str, Dict[str, float]]):
        """This function maps the virtual state of a robot to the solution of a casadi solver.
        Args:
            virtual_state ( Dict(str,Dict(str,float))): A virtual state of the robot

        Returns:
            [type]: A solution of a nlp solver
        """
        solver_state = []
        for key in self._symbolic_keys:
            solver_state.append(
                virtual_state[key[0]][key[1]])
        return solver_state
